// Program that computes the Page f(M) factor of a Schwarzschild Black Hole
// Author : Jérémy Auffinger, jeremy.auffinger@ens-lyon.fr

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>

int nb_particles = 15;
int grav = 1;
double pi = 3.14159265389795;
double Mp = 1.221e+19; // Planck mass in GeV
int nba = 50;
int nbx = 200;
int nb_masses = 100;
#define G 1./pow(1.221e+19,2.) // Newton constant in GeV
double mass_conversion = 5.60958884e+23;
double time_conversion = 1.519267407e+24;
double Mmax = 1.e+46;

double rplus_BH(double M,double a){
	// This function computes the external Kerr radius
	// of a BH of mass M and spin parameter a.
	
	return M*(1. + sqrt(1. - pow(a,2.)));
};

double temp_BH(double M,double a){
	// This function computes the Hawking temperature of a Schwarzschild
	// PBH of mass M and spin parameter a.
	
	return 1./(2.*pi*G)*(rplus_BH(M,a) - M)/(pow(rplus_BH(M,a),2.) + pow(a*M,2.));
};

void read_gamma_tables(double ***gammas,double *gamma_a,double *gamma_x){
	// This function reads the greybody factors tables.
	
	char **gamma_names = (char **)malloc(4*sizeof(char *));
	for(int i = 0;i<4;i++){
		gamma_names[i] = (char *)malloc(64*sizeof(char));
	};
	sprintf(gamma_names[0],"./spin_0_fM.txt");
	sprintf(gamma_names[1],"./spin_1_fM.txt");
	sprintf(gamma_names[2],"./spin_2_fM.txt");
	sprintf(gamma_names[3],"./spin_0.5_fM.txt");
	
	char dummy[32];
	FILE *table;
	for(int i = 0;i<4;i++){
		table = fopen(gamma_names[i],"r");
		rewind(table);
		if(!table){
			printf("\n[read_simple_tables] : ERROR COULD NOT FIND TABLE %s !\n",gamma_names[i]);
			fflush(stdout);
			return;
		};
		fscanf(table,"%s",dummy);
		for(int j = 0;j<nbx;j++){
			fscanf(table,"%lf",&(gamma_x[j]));
		};
		for(int j = 0;j<nba;j++){
			fscanf(table,"%lf",&(gamma_a[j]));
			for(int k = 0;k<nbx;k++){
				fscanf(table,"%lf",&(gammas[i][j][k]));
			};
		};
		fclose(table);
	};
	for(int i = 0;i<4;i++){
		free(gamma_names[i]);
	};
	return;
};

void read_asymp_fits(double ***fits){
	// This function reads the tables fitting asymptotic Kerr emissivities.
	
	char **table_names;
	table_names = (char **)malloc(4*sizeof(char *));
	for(int i = 0;i<4;i++){
		table_names[i] = (char *)malloc(64*sizeof(char));
	};
	sprintf(table_names[0],"./spin_0_fits_fM.txt");
	sprintf(table_names[1],"./spin_1_fits_fM.txt");
	sprintf(table_names[2],"./spin_2_fits_fM.txt");
	sprintf(table_names[3],"./spin_0.5_fits_fM.txt");
	
	FILE *file;
	char dummy[32];
	for(int i = 0;i<4;i++){
		file = fopen(table_names[i],"r");
		if(file == 0){
			printf("[read_asymp_fits] : ERROR COULD NOT OPEN FILE %s !\n",table_names[i]);
			fflush(stdout);
			return;
		};
		rewind(file);
		for(int j = 0;j<8;j++){
			fscanf(file,"%s",dummy);
		};
		for(int j = 0;j<nba;j++){
			fscanf(file,"%s",dummy);
			for(int k = 0;k<7;k++){
				fscanf(file,"%lf",&(fits[i][j][k]));
			};
		};
		fclose(file);
	};
	for(int i = 0;i<4;i++){
		free(table_names[i]);
	};
	return;
};

double nb_spin_0(double E,double M,double a,double *ms){
	// This function computes the number of spin 0 dof.
	
	double result = 0.;
	if(E > ms[5] && temp_BH(M,a) > ms[5]){ // Higgs contribution
		result += 1.; // Higgs boson (1 helicity state)
	};
	return result;
};

double nb_spin_1(double E,double M,double a,double *ms){
	// This function computes the number of spin 1 dof.
	
	double result = 2.; // photon contribution (2 helicity states)
	if(E > ms[13] && temp_BH(M,a) > ms[13]){ // W+- boson contribution
		result += 6.; // W+- bosons (2 particle species)*(3 helicity states)
	};
	if(E > ms[14] && temp_BH(M,a) > ms[14]){ // Z0 boson contribution
		result += 3.; // Z0 boson (3 helicity states)
	};
	if(E > ms[4] && temp_BH(M,a) > ms[4]){ // gluons contribution
		result += 16.; // gluons (8 particle species)*(2 helicity states)
	};
	return result;
};

double nb_spin_2(double E,double M,double a,double *ms){
	// This function computes the number of spin 2 dof.
	
	double result = 2.; // graviton (2 polarization states)
	return grav*result;
};

double nb_spin_05(double E,double M,double a,double *ms){
	// This function computes the number of spin 2 dof.
	
	double result = 6.; // neutrinos (3 particle species)*(1 helicity state)*(2 for antiparticles)
	if(E > ms[0] && temp_BH(M,a) > ms[0]){ // bottom quark contribution
		result += 12.; // bottom quark (2 helicity states)*(3 color states)*(2 for antiparticles)
	};
	if(E > ms[1] && temp_BH(M,a) > ms[1]){ // charm quark contribution
		result += 12.; // charm quark (2 helicity states)*(3 color states)*(2 for antiparticles)
	};
	if(E > ms[2] && temp_BH(M,a) > ms[2]){ // down quark contribution
		result += 12.; // down quark (2 helicity states)*(3 color states)*(2 for antiparticles)
	};
	if(E > ms[3] && temp_BH(M,a) > ms[3]){ // electron contribution
		result += 4.; // electrons (2 helicity states)*(2 for antiparticles)
	};
	if(E > ms[6] && temp_BH(M,a) > ms[6]){ // muon contribution
		result += 4.; // muon (2 helicity states)*(2 for antiparticles)
	};
	if(E > ms[9] && temp_BH(M,a) > ms[9]){ // strange quark contribution
		result += 12.; // strange quark (2 helicity states)*(3 color states)*(2 for antiparticles)
	};
	if(E > ms[10] && temp_BH(M,a) > ms[10]){ // tau contribution
		result += 4.; // tau (2 helicity states)*(2 for antiparticles)
	};
	if(E > ms[11] && temp_BH(M,a) > ms[11]){ // top quark contribution
		result += 12.; // top quark (2 helicity states)*(3 color states)*(2 for antiparticles)
	};
	if(E > ms[12] && temp_BH(M,a) > ms[12]){ // up quark contribution
		result += 12.; // up quark (2 helicity states)*(3 color states)*(2 for antiparticles)
	};
	return result;
};

double sum_gammas(double M,double a,double E,double ***gammas,int i,double *gamma_x,double ***fits,double *ms){
	// This function computes the sum over the dof of the Kerr emissivity.
	
	double x = 2.*E*M*G;
	double result = 0.;
	if(x <= 0.01){ // low-energy fit
		result += nb_spin_0(E,M,a,ms)*pow(10.,fits[0][i][0]*log10(2.*E*M*G) + fits[0][i][1])/(2.*pi); // spin 0 particles
		result += nb_spin_1(E,M,a,ms)*pow(10.,fits[1][i][0]*log10(2.*E*M*G) + fits[1][i][1])/(2.*pi); // spin 1 particles
		result += nb_spin_2(E,M,a,ms)*pow(10.,fits[2][i][0]*log10(2.*E*M*G) + fits[2][i][1])/(2.*pi); // spin 2 particles
		result += nb_spin_05(E,M,a,ms)*pow(10.,fits[3][i][0]*log10(2.*E*M*G) + fits[3][i][1])/(2.*pi); // spin 1/2 particles
	}
	else if(x >= 5.){ // high-energy fit
		result += nb_spin_0(E,M,a,ms)*pow(10.,fits[0][i][2]*(2.*M*E*G) + fits[0][i][3] + fits[0][i][4]*cos(fits[0][i][6]*2.*E*M*G) + fits[0][i][5]*sin(fits[0][i][6]*2.*E*M*G))/(2.*pi); // spin 0 particles
		result += nb_spin_1(E,M,a,ms)*pow(10.,fits[1][i][2]*(2.*M*E*G) + fits[1][i][3] + fits[1][i][4]*cos(fits[1][i][6]*2.*E*M*G) + fits[1][i][5]*sin(fits[1][i][6]*2.*E*M*G))/(2.*pi); // spin 1 particles
		result += nb_spin_2(E,M,a,ms)*pow(10.,fits[2][i][2]*(2.*M*E*G) + fits[2][i][3] + fits[2][i][4]*cos(fits[2][i][6]*2.*E*M*G) + fits[2][i][5]*sin(fits[2][i][6]*2.*E*M*G))/(2.*pi); // spin 2 particles
		result += nb_spin_05(E,M,a,ms)*pow(10.,fits[3][i][2]*(2.*M*E*G) + fits[3][i][3] + fits[3][i][4]*cos(fits[3][i][6]*2.*E*M*G) + fits[3][i][5]*sin(fits[3][i][6]*2.*E*M*G))/(2.*pi); // spin 3 particles
	}
	else{ // tabulated value
		int counter_x = 0;
		while(counter_x < nbx - 1 && 2.*E*M*G >= gamma_x[counter_x]){
			counter_x++;
		};
		result += nb_spin_0(E,M,a,ms)*(gammas[0][i][counter_x-1] + (gammas[0][i][counter_x] - gammas[0][i][counter_x-1])/(gamma_x[counter_x] - gamma_x[counter_x-1])*(2.*E*M*G - gamma_x[counter_x-1]))/(2.*pi); // spin 0 particles
		result += nb_spin_1(E,M,a,ms)*(gammas[1][i][counter_x-1] + (gammas[1][i][counter_x] - gammas[1][i][counter_x-1])/(gamma_x[counter_x] - gamma_x[counter_x-1])*(2.*E*M*G - gamma_x[counter_x-1]))/(2.*pi); // spin 1 particles
		result += nb_spin_2(E,M,a,ms)*(gammas[2][i][counter_x-1] + (gammas[2][i][counter_x] - gammas[2][i][counter_x-1])/(gamma_x[counter_x] - gamma_x[counter_x-1])*(2.*E*M*G - gamma_x[counter_x-1]))/(2.*pi); // spin 2 particles
		result += nb_spin_05(E,M,a,ms)*(gammas[3][i][counter_x-1] + (gammas[3][i][counter_x] - gammas[3][i][counter_x-1])/(gamma_x[counter_x] - gamma_x[counter_x-1])*(2.*E*M*G - gamma_x[counter_x-1]))/(2.*pi); // spin 1/2 particles
	};
	return result;
};

void compute_fM(double **fM,double *masses,double ***gammas,double *gamma_a,double *gamma_x,double ***fits,double *ms){
	// This function computes the value of f(M,a).
	
	double *energies;
	int nb_energies = 100;
	energies = (double *)malloc(nb_energies*sizeof(double));
	for(int i = 0;i<nba;i++){
		for(int j = 0;j<nb_masses;j++){
			for(int k = 0;k<nb_energies;k++){
				energies[k] = pow(10.,log10(temp_BH(masses[j],gamma_a[i])*1.e-5) + (log10(temp_BH(masses[j],gamma_a[i])*1.e+5) - log10(temp_BH(masses[j],gamma_a[i])*1.e-5))/(nb_energies - 1)*k);
			};
			for(int k = 0;k<nb_energies-1;k++){
				fM[j][i] += (energies[k+1] - energies[k])*energies[k]*pow(masses[j],2.)*sum_gammas(masses[j],gamma_a[i],energies[k],gammas,i,gamma_x,fits,ms);
			};
		};
	};
	free(energies);
	return;
};

void write_fM(double **fM,double *masses,double *gamma_a){
	// This function writes the f(M,a) tables.
	
	FILE *file = fopen("fM.txt","w+");
	if(!file){
		printf("\n Could not open file fM.txt\n");
		fflush(stdout);
		return;
	};
	rewind(file);
	fprintf(file,"%15s","mass/a");
	for(int i = 0;i<nba;i++){
		fprintf(file,"%15.5e",gamma_a[i]);
	};
	fprintf(file,"\n");
	for(int i = 0;i<nb_masses;i++){
		fprintf(file,"%15.5e",masses[i]);
		for(int j = 0;j<nba;j++){
			fprintf(file,"%15.5e",fM[i][j]);
		};
		fprintf(file,"\n");
	};
	fclose(file);
	return;
};

void read_gamma_tables_2(double ***gammas,double *gamma_a,double *gamma_x){
	// This function reads the m*greybody factors tables.
	
	char **gamma_names = (char **)malloc(4*sizeof(char *));
	for(int i = 0;i<4;i++){
		gamma_names[i] = (char *)malloc(64*sizeof(char));
	};
	sprintf(gamma_names[0],"./spin_0_gM.txt");
	sprintf(gamma_names[1],"./spin_1_gM.txt");
	sprintf(gamma_names[2],"./spin_2_gM.txt");
	sprintf(gamma_names[3],"./spin_0.5_gM.txt");
	
	char dummy[32];
	FILE *table;
	for(int i = 0;i<4;i++){
		table = fopen(gamma_names[i],"r");
		rewind(table);
		if(!table){
			printf("\n[read_simple_tables] : ERROR COULD NOT FIND TABLE %s !\n",gamma_names[i]);
			fflush(stdout);
			return;
		};
		fscanf(table,"%s",dummy);
		for(int j = 0;j<nbx;j++){
			fscanf(table,"%lf",&(gamma_x[j]));
		};
		for(int j = 0;j<nba;j++){
			fscanf(table,"%lf",&(gamma_a[j]));
			for(int k = 0;k<nbx;k++){
				fscanf(table,"%lf",&(gammas[i][j][k]));
			};
		};
		fclose(table);
	};
	for(int i = 0;i<4;i++){
		free(gamma_names[i]);
	};
	return;
};

void read_asymp_fits_2(double ***fits){
	// This function reads the tables fitting m*asymptotic Kerr emissivities.
	
	char **table_names;
	table_names = (char **)malloc(4*sizeof(char *));
	for(int i = 0;i<4;i++){
		table_names[i] = (char *)malloc(64*sizeof(char));
	};
	sprintf(table_names[0],"./spin_0_fits_gM.txt");
	sprintf(table_names[1],"./spin_1_fits_gM.txt");
	sprintf(table_names[2],"./spin_2_fits_gM.txt");
	sprintf(table_names[3],"./spin_0.5_fits_gM.txt");
	
	FILE *file;
	char dummy[32];
	for(int i = 0;i<4;i++){
		file = fopen(table_names[i],"r");
		if(file == 0){
			printf("[read_asymp_fits] : ERROR COULD NOT OPEN FILE %s !\n",table_names[i]);
			fflush(stdout);
			return;
		};
		rewind(file);
		for(int j = 0;j<8;j++){
			fscanf(file,"%s",dummy);
		};
		for(int j = 0;j<nba;j++){
			fscanf(file,"%s",dummy);
			for(int k = 0;k<7;k++){
				fscanf(file,"%lf",&(fits[i][j][k]));
			};
		};
		fclose(file);
	};
	for(int i = 0;i<4;i++){
		free(table_names[i]);
	};
	return;
};

void compute_gM(double **gM,double *masses,double ***gammas,double *gamma_a,double *gamma_x,double ***fits,double *ms){
	// This function computes the value of g(M,a).
	
	double *energies;
	int nb_energies = 100;
	energies = (double *)malloc(nb_energies*sizeof(double));
	for(int i = 0;i<nba;i++){
		for(int j = 0;j<nb_masses;j++){
			for(int k = 0;k<nb_energies;k++){
				energies[k] = pow(10.,log10(temp_BH(masses[j],gamma_a[i])*1.e-5) + (log10(temp_BH(masses[j],gamma_a[i])*1.e+5) - log10(temp_BH(masses[j],gamma_a[i])*1.e-5))/(nb_energies - 1)*k);
			};
			for(int k = 0;k<nb_energies-1;k++){
				if(gamma_a[i] != 0.){
					gM[j][i] += (energies[k+1] - energies[k])*masses[j]/gamma_a[i]*sum_gammas(masses[j],gamma_a[i],energies[k],gammas,i,gamma_x,fits,ms);
				};
			};
		};
	};
	free(energies);
	return;
};

void write_gM(double **gM,double *masses,double *gamma_a){
	// This function writes the g(M,a) tables.
	
	FILE *file = fopen("gM.txt","w+");
	if(!file){
		printf("\n Could not open file gM.txt\n");
		fflush(stdout);
		return;
	};
	rewind(file);
	fprintf(file,"%15s","mass/a");
	for(int i = 0;i<nba;i++){
		fprintf(file,"%15.5e",gamma_a[i]);
	};
	fprintf(file,"\n");
	for(int i = 0;i<nb_masses;i++){
		fprintf(file,"%15.5e",masses[i]);
		for(int j = 0;j<nba;j++){
			fprintf(file,"%15.5e",gM[i][j]/G);
		};
		fprintf(file,"\n");
	};
	fclose(file);
	return;
};

int main(){
	
	double ms[nb_particles+1];
	ms[0] = 4.18;
	ms[1] = 1.27;
	ms[2] = 4.7e-3;
	ms[3] = 0.5109989461e-3;
	ms[4] = 200.e-3; // effective mass
	ms[5] = 125.03;
	ms[6] = 105.6583745e-3;
	ms[7] = 0.;
	ms[8] = 0.;
	ms[9] = 96e-3;
	ms[10] = 1.77686;
	ms[11] = 173.21;
	ms[12] = 2.2e-3;
	ms[13] = 80.403;
	ms[14] = 91.1876;
	ms[15] = 0.;
	
	printf(" Reading gamma tables...");
	fflush(stdout);
	
	double *gamma_a;
	gamma_a = (double *)malloc(nba*sizeof(double));
	double *gamma_x;
	gamma_x = (double *)malloc(nbx*sizeof(double));
	double ***gammas;
	gammas = (double ***)malloc(4*sizeof(double **));
	for(int i = 0;i<4;i++){
		gammas[i] = (double **)malloc(nba*sizeof(double *));
		for(int j = 0;j<nba;j++){
			gammas[i][j] = (double *)malloc(nbx*sizeof(double));
		};
	};
	
	read_gamma_tables(gammas,gamma_a,gamma_x);
	
	printf("\t DONE");
	printf("\n Reading asymptotical fits...");
	fflush(stdout);
	
	double ***fits;
	fits = (double ***)malloc(4*sizeof(double **));
	for(int i = 0;i<4;i++){
		fits[i] = (double **)malloc(nba*sizeof(double *));
		for(int j = 0;j<nba;j++){
			fits[i][j] = (double *)malloc(7*sizeof(double));
		};
	};
	
	read_asymp_fits(fits);
	
	printf("\t DONE");
	printf("\n Computing f(M,a) table...");
	fflush(stdout);
	
	double **fM;
	fM = (double **)malloc(nb_masses*sizeof(double *));
	for(int i = 0;i<nb_masses;i++){
		fM[i] = (double *)malloc(nba*sizeof(double));
		for(int j = 0;j<nba;j++){
			fM[i][j] = 0.;
		};
	};
	double *masses;
	masses = (double *)malloc(nb_masses*sizeof(double));
	for(int i = 0;i<nb_masses;i++){
		masses[i] = pow(10.,log10(Mp) + (log10(Mmax) - log10(Mp))/(nb_masses - 1)*i);
	};
	
	compute_fM(fM,masses,gammas,gamma_a,gamma_x,fits,ms);
	
	printf("\t DONE");
	printf("\n Writing f(M,a) table...");
	
	write_fM(fM,masses,gamma_a);
	
	for(int i = 0;i<nb_masses;i++){
		free(fM[i]);
	};
	
	printf("\t DONE");
	printf("\n Reading m*gamma tables...");
	fflush(stdout);
	
	read_gamma_tables_2(gammas,gamma_a,gamma_x);
	
	printf("\t DONE");
	printf("\n Reading m*fits...");
	fflush(stdout);
	
	read_asymp_fits_2(fits);
	
	printf("\t\t DONE");
	printf("\n Computing g(M,a) table...");
	fflush(stdout);
	
	double **gM;
	gM = (double **)malloc(nb_masses*sizeof(double *));
	for(int i = 0;i<nb_masses;i++){
		gM[i] = (double *)malloc(nba*sizeof(double));
		for(int j = 0;j<nba;j++){
			gM[i][j] = 0.;
		};
	};
	
	compute_gM(gM,masses,gammas,gamma_a,gamma_x,fits,ms);
	
	printf("\t DONE");
	printf("\n Writing g(M,a) table...");
	fflush(stdout);
	
	write_gM(gM,masses,gamma_a);
	
	printf("\t DONE\n");
	fflush(stdout);
	
	for(int i = 0;i<nb_masses;i++){
		free(gM[i]);
	};
	for(int i = 0;i<4;i++){
		for(int j = 0;j<nba;j++){
			free(fits[i][j]);
			free(gammas[i][j]);
		};
	};
	free(masses);
	free(gamma_a);
	free(gamma_x);
	free(ms);
	
	return 1;
};










