// This is the source file where the methods computing
// the time evolution of Black Holes are implemented.
// Last modification: 14 November 2019
// Authors: Jérémy Auffinger j.auffinger@ipnl.in2p3.fr & Alexandre Arbey alexandre.arbey@ens-lyon.fr

#include "include.h"

double rplus_BH(double M,double a){
	// This function computes the external Kerr radius
	// of a BH of mass M and spin parameter a.
	
	return M*(1. + sqrt(1. - pow(a,2.)));
}

double temp_BH(double M,double a){
	// This function computes the Hawking temperature of a Schwarzschild
	// PBH of mass M and spin parameter a.
	
	return 1./(2.*pi*G)*(rplus_BH(M,a) - M)/(pow(rplus_BH(M,a),2.) + pow(a*M,2.));
}

void read_fM_table(double **fM_table,double *fM_masses,double *fM_a,struct param *parameters){
	// This function reads the f(M,a) factor table used to compute the evolution
	// of BH, stored in the folder /fM_tables.
	
	FILE *table_file;
	char name[64];
	switch(parameters->grav){
		case 0:{ // we use the f(M,a) table computed without the graviton
			sprintf(name,"./src/tables/fM_tables/fM.txt"); // for the moment only with the graviton
			break;
		}
		case 1:{ // we use the f(M,a) table computed taking the graviton into account
			sprintf(name,"./src/tables/fM_tables/fM_nograv.txt");
			break;
		}
		default:{
			printf("\n\t [read_fM_table] : ERROR WRONG GRAVITON CHOICE !\n");
			fflush(stdout);
			exit(0);
			break;
		}
	}
	table_file = fopen(name,"r");
	if(!table_file){
		printf("\n\t [read_fM_table] : ERROR COULD NOT READ TABLE '%s' !\n",name);
		fflush(stdout);
		exit(0);
	}
	rewind(table_file);
	if(parameters->full_output){
		printf("\n\n\t reading:\t %s\n",name);
		fflush(stdout);
	}
	char dummy[20];
	fscanf(table_file,"%s",dummy);
	for(int i = 0;i<parameters->nb_fM_a;i++){
		fscanf(table_file,"%lf",&(fM_a[i]));
	}
	for(int i = 0;i<parameters->nb_fM_masses;i++){
		fscanf(table_file,"%lf",&(fM_masses[i]));
		for(int j = 0;j<parameters->nb_fM_a;j++){
			fscanf(table_file,"%lf",&(fM_table[i][j]));
		}
	}
	fclose(table_file);
	return;
}

void read_gM_table(double **gM_table,double *fM_masses,double *fM_a,struct param *parameters){
	// This function reads the g(M,a) factor table used to compute the evolution
	// of BH, stored in the folder /fM_tables.
	
	FILE *table_file;
	char name[64];
	switch(parameters->grav){
		case 0:{ // we use the g(M,a) table computed without the graviton
			sprintf(name,"./src/tables/fM_tables/gM.txt"); // for the moment only with the graviton
			break;
		}
		case 1:{ // we use the g(M,a) table computed taking the graviton into account
			sprintf(name,"./src/tables/fM_tables/gM_nograv.txt");
			break;
		}
		default:{
			printf("\n\t [read_gM_table] : ERROR WRONG GRAVITON CHOICE !\n");
			fflush(stdout);
			exit(0);
			break;
		}
	}
	table_file = fopen(name,"r");
	if(!table_file){
		printf("\n\t [read_gM_table] : ERROR COULD NOT READ TABLE '%s' !\n",name);
		fflush(stdout);
		exit(0);
	}
	rewind(table_file);
	if(parameters->full_output){
		printf("\t\t\t %s\n\n",name);
		fflush(stdout);
	}
	char dummy[20];
	fscanf(table_file,"%s",dummy);
	for(int i = 0;i<parameters->nb_fM_a;i++){
		fscanf(table_file,"%lf",&(fM_a[i]));
	}
	for(int i = 0;i<parameters->nb_fM_masses;i++){
		fscanf(table_file,"%lf",&(fM_masses[i]));
		for(int j = 0;j<parameters->nb_fM_a;j++){
			fscanf(table_file,"%lf",&(gM_table[i][j]));
		}
	}
	fclose(table_file);
	return;
}

double loss_rate_M(double M,double a,double **fM_table,double *fM_masses,double *fM_a,int counter_M,int counter_a,struct param *parameters){
	// This function computes the BH mass loss rate dM/dt by using the f(M,a) table.
	
	double fM;
	if(parameters->interpolation_method == 0){ // linear interpolation
		fM = fM_table[counter_M][counter_a] + (fM_table[counter_M+1][counter_a] - fM_table[counter_M][counter_a])/(fM_masses[counter_M+1] - fM_masses[counter_M])*(M - fM_masses[counter_M]) + (fM_table[counter_M][counter_a+1] - fM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
	}
	else if(parameters->interpolation_method == 1){ // logarithmic interpolation
		fM = pow(10.,log10(fM_table[counter_M][counter_a]) + (log10(fM_table[counter_M+1][counter_a]) - log10(fM_table[counter_M][counter_a]))/(log10(fM_masses[counter_M+1]) - log10(fM_masses[counter_M]))*(log10(M) - log10(fM_masses[counter_M]))) + (fM_table[counter_M][counter_a+1] - fM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
	}
	else{
		printf("\n\t [loss_rate_M] : ERROR WRONG INTERPOLATION METHOD !\n");
		fflush(stdout);
		exit(1);
	}
	return -fM/pow(M,2.);
}

double loss_rate_a(double M,double a,double **fM_table,double **gM_table,double *fM_masses,double *fM_a,int counter_M,int counter_a,struct param *parameters){
	// This function computes the BH spin loss rate da/dt by using the f(M,a) and g(M,a) tables.
	
	if(a == 0.){ // no more evolution of the spin
		return 0.;
	}
	else{
		double fM;
		double gM;
		if(parameters->interpolation_method == 0 || counter_a == 0){ // linear interpolation
			fM = fM_table[counter_M][counter_a] + (fM_table[counter_M+1][counter_a] - fM_table[counter_M][counter_a])/(fM_masses[counter_M+1] - fM_masses[counter_M])*(M - fM_masses[counter_M]) + (fM_table[counter_M][counter_a+1] - fM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
			gM = gM_table[counter_M][counter_a] + (gM_table[counter_M+1][counter_a] - gM_table[counter_M][counter_a])/(fM_masses[counter_M+1] - fM_masses[counter_M])*(M - fM_masses[counter_M]) + (gM_table[counter_M][counter_a+1] - gM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
		}
		else if(parameters->interpolation_method == 1){ // logarithmic interpolation
			fM = pow(10.,log10(fM_table[counter_M][counter_a]) + (log10(fM_table[counter_M+1][counter_a]) - log10(fM_table[counter_M][counter_a]))/(log10(fM_masses[counter_M+1]) - log10(fM_masses[counter_M]))*(log10(M) - log10(fM_masses[counter_M]))) + (fM_table[counter_M][counter_a+1] - fM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
			gM = pow(10.,log10(gM_table[counter_M][counter_a]) + (log10(gM_table[counter_M+1][counter_a]) - log10(gM_table[counter_M][counter_a]))/(log10(fM_masses[counter_M+1]) - log10(fM_masses[counter_M]))*(log10(M) - log10(fM_masses[counter_M]))) + (gM_table[counter_M][counter_a+1] - gM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
		}
		else{
			printf("\n\t [loss_rate_a] : ERROR WRONG INTERPOLATION METHOD !\n");
			fflush(stdout);
			exit(0);
		}
		return -gM*a/pow(M,3.) + 2.*fM*a/pow(M,3.);
	}
}


void evolution_times(double *init_masses,double *init_spins,double **evol_times,double **fM_table,double **gM_table,double *fM_masses,double *fM_a,
	struct param *parameters){
	// This function computes the lifetimes of each BH in the spec_table grid, in order to
	// sort them and predict the correct integration order for life_evolution(). The limit
	// parameter should not be reached.
	
	double life_mass; // running BH mass
	double life_spin; // running BH spin
	double dt; // running time step
	int counter_M; // running position in f(M,a) / g(M,a) tables
	int counter_a; // running position in f(M,a) / g(M,a) tables
	double estimate_M; // intermediate evaluations
	double estimate_a; // intermediate evaluations
	int counter; // number of time steps for each mass, should not reach the limit parameter
	double evol_M; // relative BH mass evolution
	double evol_a; // relative BH spin evolution
	
	for(int i = 0;i<parameters->BHnumber;i++){
		for(int j = 0;j<parameters->anumber;j++){
			life_mass = init_masses[i]; // initialization of the mass
			life_spin = init_spins[j]; // initialization of the spin
			dt = parameters->tmin; // initialization of the timestep
			evol_times[i][j] = parameters->tmin; // initialization of the lifetime
			counter_M = parameters->nb_fM_masses-1; // the BH mass is decreasing
			counter_a = parameters->nb_fM_a-1; // the BH spin is decreasing
			counter = 0;
			while(counter < parameters->limit-1 && life_mass > 1.1*Mp){ // recursion limit, evaporation until Planck mass
				while(counter_M > 0 && life_mass < fM_masses[counter_M]){ // we find the closest tabulated value
					counter_M--;
				}
				while(counter_a > 0 && life_spin < fM_a[counter_a]){ // we find the closest tabulated value
					counter_a--;
				}
				
				estimate_M = life_mass + dt*loss_rate_M(life_mass,life_spin,fM_table,fM_masses,fM_a,counter_M,counter_a,parameters); // Euler estimation of the next mass
				estimate_a = life_spin + dt*loss_rate_a(life_mass,life_spin,fM_table,gM_table,fM_masses,fM_a,counter_M,counter_a,parameters); // Euler estimation of the next spin
				evol_M = estimate_M/life_mass;
				if(life_spin != 0.){
					evol_a = estimate_a/life_spin;
				}
				else{
					evol_a = 0.;
				}
				
				if(evol_M < 0.9 || (evol_a != 0. && evol_a < 0.9)){ // one of the variations is too high, we diminish the time step
					dt = dt/2.;
				}
				else if(evol_M > 0.999 && (evol_a == 0. || evol_a > 0.999) && dt/evol_times[i][j] < 1.){ // all the variations are too low and the time step is not too high compared to the time scale, we increase the time step
					dt = dt*2.;
				}
				else{ // all the variations are fine, we record the evolution and move forward
					evol_times[i][j] = evol_times[i][j] + dt;
					life_mass = estimate_M;
					if(estimate_a < 1.e-3){ // we stop evolving the spin and approximate it to 0
						life_spin = 0.;
					}
					else{
						life_spin = estimate_a;
					}
				}
				
				counter++;
			}
			if(counter >= parameters->limit){
				printf("\n\t [evolution_times] : ERROR ITERATION LIMIT REACHED !\n");
				fflush(stdout);
				exit(0);
			}
		}
	}
	
	/*char file_name[500];
	sprintf(file_name,"./results/%s/evolutions_times.txt",parameters->destination_folder);
	FILE *file = fopen(file_name,"w");
	if(!file){
		printf("\n\t [evolution_times] : ERROR COULD NOT OPEN FILE '%s' !\n",file_name);
		fflush(stdout);
		return;
	}
	rewind(file);
	fprintf(file,"%15s","mass/spin");
	for(int i = 0;i<parameters->anumber;i++){
		fprintf(file,"%15.5f",init_spins[i]);
	}
	fprintf(file,"\n");
	for(int i = 0;i<parameters->BHnumber;i++){
		fprintf(file,"%15.5e",init_masses[i]/mass_conversion);
		for(int j = 0;j<parameters->anumber;j++){
			fprintf(file,"%15.5e",evol_times[i][j]/time_conversion);
		}
		fprintf(file,"\n");
	}
	fclose(file);*/
	
	return;
}

void sort_lifetimes(double **evol_times,double *sorted_times,int **rank,struct param *parameters){
	// This function sorts the lifetime of BHs to use the simple algorithm of
	// life_evolutions().
	
	for(int i = 0;i<parameters->BHnumber;i++){
		for(int j = 0;j<parameters->anumber;j++){
			sorted_times[parameters->BHnumber*i+j] = evol_times[i][j];
		}
	}
	sort_fusion(sorted_times,parameters->BHnumber*parameters->anumber);
	int counter_rank;
	for(int i = 0;i<parameters->BHnumber;i++){
		for(int j = 0;j<parameters->anumber;j++){
			counter_rank = 0;
			while(counter_rank < parameters->BHnumber*parameters->anumber && fabs(evol_times[i][j] - sorted_times[counter_rank])/sorted_times[counter_rank] > machine_precision){
				counter_rank++;
			}
			rank[i][j] = counter_rank;
		}
	}
	
	return;
}

void life_evolution(double ***life_masses,double ***life_spins,double *life_times,double *dts,double *init_masses,
	double *init_spins,int **rank,double **fM_table,double **gM_table,double *fM_masses,double *fM_a,struct param *parameters){
	// This function computes the evolution of BH from their initial mass to the Planck mass.
	// It uses the function loss_rate to compute as estimation of M(t+dt) and a(t+dt) through a Euler
	// method, and adapts the time step to keep relative variations between 0.001 and 0.1.
	// The intermediate masses, spins, times and time intervals are stored in the arrays life_masses[][][],
	// life_spins[][][], life_times[] and dts[] respectively. A recursion limit is set by the limit parameter
	// but should not be reached.
	
	double dt = parameters->tmin; // running time step
	int current_ind = 0; // running index of computation
	int **counters_M = (int **)malloc(parameters->BHnumber*sizeof(int *)); // positions in the f(M,a) / g(M,a) tables
	int **counters_a = (int **)malloc(parameters->BHnumber*sizeof(int *)); // positions in the f(M,a) / g(M,a) tables
	double **estimates_M = (double **)malloc(parameters->BHnumber*sizeof(double *)); // Euler estimates
	double **estimates_a = (double **)malloc(parameters->BHnumber*sizeof(double *)); // Euler estimates
	life_times[0] = parameters->tmin;
	dts[0] = parameters->tmin;
	for(int i = 0;i<parameters->BHnumber;i++){
		estimates_M[i] = (double *)malloc(parameters->anumber*sizeof(double));
		estimates_a[i] = (double *)malloc(parameters->anumber*sizeof(double));
		counters_M[i] = (int *)malloc(parameters->anumber*sizeof(int));
		counters_a[i] = (int *)malloc(parameters->anumber*sizeof(int));
		for(int j = 0;j<parameters->anumber;j++){
			life_masses[i][j][0] = init_masses[i]; // initialization of the masses
			life_spins[i][j][0] = init_spins[j]; // initialization of the spins
			counters_M[i][j] = parameters->nb_fM_masses-1; // masses are decreasing
			counters_a[i][j] = parameters->nb_fM_a-1; // spins are decreasing
		}
	}
	double ratio_M; // contains the relative evolution of the BH mass
	double ratio_a; // contains the relative evolution of the BH spin
	int counter = 0; // contains the current total number of tried steps
	int interesting_BH = 0; // contains the running index in the sorted lifetimes
	int interesting_i; // contains the position of the interesting BH mass
	int interesting_j; // contains the position of the interesting BH spin
	int output_BH = 0; // decides whether we print the computation advancement
	parameters->nb_fin_times = 0;
	if(parameters->full_output){
		printf("\n\n\t computing:");
		fflush(stdout);
	}
	while(counter<parameters->limit-1 && interesting_BH < parameters->BHnumber*parameters->anumber){ // limit of recursion
#if defined(_OPENMP)
#pragma omp parallel for
#endif
		for(int i = 0;i<parameters->BHnumber;i++){
			for(int j = 0;j<parameters->anumber;j++){
				if(life_masses[i][j][current_ind] > 1.1*Mp){ // We continue to evolve this BH
					while(counters_M[i][j] > 0 && life_masses[i][j][current_ind] < fM_masses[counters_M[i][j]]){ // we find the closest tabulated value
						counters_M[i][j]--;
					}
					while(counters_a[i][j] > 0 && life_spins[i][j][current_ind] < fM_a[counters_a[i][j]]){ // we find the closest tabulated value
						counters_a[i][j]--;
					}
					estimates_M[i][j] = life_masses[i][j][current_ind] + dt*loss_rate_M(life_masses[i][j][current_ind],life_spins[i][j][current_ind],fM_table,fM_masses,fM_a,counters_M[i][j],counters_a[i][j],parameters); // Euler estimation of the next mass
					estimates_a[i][j] = life_spins[i][j][current_ind] + dt*loss_rate_a(life_masses[i][j][current_ind],life_spins[i][j][current_ind],fM_table,gM_table,fM_masses,fM_a,counters_M[i][j],counters_a[i][j],parameters); // Euler estimation of the next spin
				}
				else{ // end of the evolution for this BH
					if(rank[i][j] >= interesting_BH){
						interesting_BH++;
					}
					estimates_M[i][j] = 0.;
					estimates_a[i][j] = 0.;
				}
			}
		}
		if(interesting_BH < parameters->BHnumber*parameters->anumber){	
			for(int i = 0;i<parameters->BHnumber;i++){
				for(int j = 0;j<parameters->anumber;j++){
					if(rank[i][j] == interesting_BH){
						interesting_i = i;
						interesting_j = j;
					}
				}
			}
			ratio_M = estimates_M[interesting_i][interesting_j]/life_masses[interesting_i][interesting_j][current_ind]; // the relative mass variation
			if(life_spins[interesting_i][interesting_j][current_ind] != 0.){
				ratio_a = estimates_a[interesting_i][interesting_j]/life_spins[interesting_i][interesting_j][current_ind]; // the relative spin variation
			}
			else{
				ratio_a = 0.;
			}
			if(ratio_M < 0.9 || (ratio_a != 0. && ratio_a < 0.9)){ // one of the variations is too high, we diminish the time step
				dt = dt/2.;
			}
			else if(ratio_M > 0.999 && (ratio_a == 0. || ratio_a > 0.999) && dt/life_times[current_ind] < 1.){ // all the variations are too low and the time step is not too high compared to the time scale, we increase the time step
				dt = dt*2.;
			}
			else{ // all the variations are fine, we record the evolution and move forward
				current_ind++;
				life_times[current_ind] = life_times[current_ind-1] + dt;
				dts[current_ind] = dt;
				parameters->nb_fin_times = current_ind;
				for(int i = 0;i<parameters->BHnumber;i++){
					for(int j = 0;j<parameters->anumber;j++){
						if(estimates_M[i][j] > Mp){
							life_masses[i][j][current_ind] = estimates_M[i][j];
							if(estimates_a[i][j] < 1.e-3){ // limit spin under which we consider that it is 0
								life_spins[i][j][current_ind] = 0.;
							}
							else{
								life_spins[i][j][current_ind] = estimates_a[i][j];
							}
						}
						else{
							life_masses[i][j][current_ind] = 0.;
							life_spins[i][j][current_ind] = 0.;
						}
					}
				}
			}
			counter++;
			if(parameters->full_output && output_BH != interesting_BH){
				output_BH = interesting_BH;
				if(output_BH%(parameters->BHnumber*parameters->anumber/10) == 0){
					printf("\t%i/%i",output_BH,parameters->BHnumber*parameters->anumber);
					fflush(stdout);
				}
			}
		}
	}
	if(parameters->full_output){
		printf("\t%i/%i\n\n",parameters->BHnumber*parameters->anumber,parameters->BHnumber*parameters->anumber);
		fflush(stdout);
	}
	free2D_int(counters_M,parameters->BHnumber);
	free2D_int(counters_a,parameters->BHnumber);
	free2D_double(estimates_M,parameters->BHnumber);
	free2D_double(estimates_a,parameters->BHnumber);
	if(counter == parameters->limit-1){
		printf("\n\t [life_evolution] : ERROR ITERATION LIMIT REACHED !\n");
		fflush(stdout);
	}
	return;
}

void write_life_evolutions(double ***life_masses,double ***life_spins,double *life_times,struct param *parameters){
	// This function writes the number of time steps and the masses and spins as
	// functions of time computed by the function life_evolution for
	// all initial BH masses and spins in the file 'life_evolutions.txt'.
	
	char file_name[500];
	sprintf(file_name,"./results/%s/life_evolutions.txt",parameters->destination_folder);
	FILE *file = fopen(file_name,"w");
	if(!file){
		printf("\n\t [write_life_evolutions] : ERROR COULD NOT OPEN FILE '%s' !\n",file_name);
		fflush(stdout);
		exit(0);
	}
	rewind(file);
	fprintf(file,"Evolution of the BH masses and spins as functions of time.\n");
	fprintf(file,"Total number of time iterations: %i\n\n",parameters->nb_fin_times);
	
	for(int i = 0;i<parameters->BHnumber;i++){
		fprintf(file,"%15s","t");
		for(int j = 0;j<parameters->anumber;j++){
			fprintf(file,"%5s%15s%15s","","M","a");
		}
		fprintf(file,"\n");
		for(int k = 0;k<parameters->nb_fin_times;k++){
			fprintf(file,"%15.5e",life_times[k]/time_conversion);
			for(int j = 0;j<parameters->anumber;j++){
				fprintf(file,"%5s%15.5e%15.5e","",life_masses[i][j][k]/mass_conversion,life_spins[i][j][k]);
			}
			fprintf(file,"\n");
		}
		fprintf(file,"\n\n");
	}
	
	fclose(file);
	return;
}
