// This is the source file where the methods computing
// the primary spectra are implemented.
// Last modification: 14 November 2019
// Authors: Jérémy Auffinger j.auffinger@ipnl.in2p3.fr & Alexandre Arbey alexandre.arbey@ens-lyon.fr

#include "include.h"

void read_gamma_tables(double ***gammas,double *gamma_a,double *gamma_x,struct param *parameters){
	// This function reads the greybody factors tables stored in the folder '/gamma_tables'.
	// It fills the tabulated BH spins, energies and factors in the arrays
	// gamma_a[], gamma_x[] and gammas[][][] respectively.
	
	char **gamma_names = (char **)malloc(4*sizeof(char *)); // contains the names of the greybody tables files
	for(int i = 0;i<4;i++){
		gamma_names[i] = (char *)malloc(64*sizeof(char));
	}
	sprintf(gamma_names[0],"./src/tables/gamma_tables/spin_0.txt");
	sprintf(gamma_names[1],"./src/tables/gamma_tables/spin_1.txt");
	sprintf(gamma_names[2],"./src/tables/gamma_tables/spin_2.txt");
	sprintf(gamma_names[3],"./src/tables/gamma_tables/spin_0.5.txt");
	
	char dummy[32];
	FILE *table;
	if(parameters->full_output){
		printf("\n\n\t reading:\t");
		fflush(stdout);
	}
	for(int i = 0;i<4;i++){
		table = fopen(gamma_names[i],"r");
		if(!table){
			printf("\n\t [read_simple_tables] : ERROR COULD NOT FIND TABLE '%s' !\n",gamma_names[i]);
			fflush(stdout);
			exit(0);
		}
		rewind(table);
		fscanf(table,"%s",dummy);
		for(int j = 0;j<parameters->nb_gamma_x;j++){
			fscanf(table,"%lf",&(gamma_x[j]));
		}
		for(int j = 0;j<parameters->nb_gamma_a;j++){
			fscanf(table,"%lf",&(gamma_a[j]));
			for(int k = 0;k<parameters->nb_gamma_x;k++){
				fscanf(table,"%lf",&(gammas[i][j][k]));
			}
		}
		fclose(table);
		if(parameters->full_output){ // prints the state of reading
			if(i == 0){
				printf("%s\n",gamma_names[i]);
			}
			else{
				printf("\t\t\t%s\n",gamma_names[i]);
			}
			fflush(stdout);
		}
	}
	if(parameters->full_output){
		printf("\n");
		fflush(stdout);
	}
	free2D_char(gamma_names,4);
	return;
}

void read_asymp_fits(double ***fits,struct param *parameters){
	// This function reads the tables fitting asymptotic Kerr emissivities
	// in the folder "/gamma_tables". It fills the array fits[][][].
	
	char **table_names;
	table_names = (char **)malloc(4*sizeof(char *)); // contains the names of the fits tables files
	for(int i = 0;i<4;i++){
		table_names[i] = (char *)malloc(64*sizeof(char));
	}
	sprintf(table_names[0],"./src/tables/gamma_tables/spin_0_fits.txt");
	sprintf(table_names[1],"./src/tables/gamma_tables/spin_1_fits.txt");
	sprintf(table_names[2],"./src/tables/gamma_tables/spin_2_fits.txt");
	sprintf(table_names[3],"./src/tables/gamma_tables/spin_0.5_fits.txt");
	
	FILE *file;
	char dummy[32];
	if(parameters->full_output){
		printf("\n\n\t reading:\t");
		fflush(stdout);
	}
	for(int i = 0;i<4;i++){
		file = fopen(table_names[i],"r");
		if(file == 0){
			printf("\n\t [read_asymp_fits] : ERROR COULD NOT OPEN FILE '%s' !\n",table_names[i]);
			fflush(stdout);
			exit(0);
		}
		rewind(file);
		for(int j = 0;j<8;j++){
			fscanf(file,"%s",dummy);
		}
		for(int j = 0;j<parameters->nb_gamma_a;j++){
			fscanf(file,"%s",dummy);
			for(int k = 0;k<7;k++){
				fscanf(file,"%lf",&(fits[i][j][k]));
			}
		}
		if(parameters->full_output){ // prints the state of reading
			if(i == 0){
				printf("%s\n",table_names[i]);
			}
			else{
				printf("\t\t\t%s\n",table_names[i]);
			}
			fflush(stdout);
		}
		fclose(file);
	}
	if(parameters->full_output){
		printf("\n");
		fflush(stdout);
	}
	free2D_char(table_names,4);
	return;
}

double dNdtdE(double E,double M,double a,int particle_index,double ***gammas,double *gamma_a,double *gamma_x,double ***fits,double *dof,double *spins,double *masses_primary,int counter_a,int counter_x,struct param *parameters){
	// This function computes the emission rate d²N/dtdE for the elementary particle of type
	// particle_index. It uses the greybody factors tables or the asymptotical fits.
	
	double T = temp_BH(M,a);
	int type; // contains the spin type of the particle
	if(spins[particle_index] == 0.){
		type = 0;
	}
	if(spins[particle_index] == 1.){
		type = 1;
	}
	if(spins[particle_index] == 2.){
		type = 2;
	}
	if(spins[particle_index] == 0.5){
		type = 3;
	}
	if(E < masses_primary[particle_index]){
		return 0.;
	}
	else if(2.*E*M*G > 0.01 && 2.*E*M*G < 5.){
		switch(parameters->interpolation_method){
			case 0:{ // linear interpolation
				return dof[particle_index]*(gammas[type][counter_a-1][counter_x-1] + (gammas[type][counter_a][counter_x-1] - gammas[type][counter_a-1][counter_x-1])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]) + (gammas[type][counter_a-1][counter_x] - gammas[type][counter_a-1][counter_x-1])/(gamma_x[counter_x] - gamma_x[counter_x-1])*(2.*E*M*G - gamma_x[counter_x-1]))/(2.*pi);
				break;
			}
			case 1:{ // logarithmic interpolation
				if(gammas[type][counter_a-1][counter_x-1] == 0. || gammas[type][counter_a][counter_x-1] == 0. || gammas[type][counter_a-1][counter_x] == 0.){
					return 0.;
				}
				else{
					return dof[particle_index]*(pow(10.,log10(gammas[type][counter_a-1][counter_x-1]) + (log10(gammas[type][counter_a-1][counter_x]) - log10(gammas[type][counter_a-1][counter_x-1]))/(log10(gamma_x[counter_x]) - log10(gamma_x[counter_x-1]))*(log10(2.*E*M*G) - log10(gamma_x[counter_x-1]))) + (gammas[type][counter_a][counter_x-1] - gammas[type][counter_a-1][counter_x-1])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]))/(2.*pi);
				}
				break;
			}
			default:{
				printf("\n\t [dNdtdE] : ERROR WRONG INTERPOLATION METHOD !\n");
				fflush(stdout);
				exit(0);
				break;
			}
		}
	}
	else if(2.*E*M*G <= 0.01){ // low-energy asymptotic value
		double a1;
		double b1;
		a1 = fits[type][counter_a-1][0] + (fits[type][counter_a][0] - fits[type][counter_a-1][0])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]);
		b1 = fits[type][counter_a-1][1] + (fits[type][counter_a][1] - fits[type][counter_a-1][1])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]);
		return dof[particle_index]*pow(10.,a1*log10(2.*E*M*G) + b1)/(2.*pi);
	}
	else{ // high-energy asymptotic value 2.*E*M*G >= 5.
		double a1;
		double b1;
		double c1;
		double d1;
		double e1;
		a1 = fits[type][counter_a-1][2] + (fits[type][counter_a][2] - fits[type][counter_a-1][2])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]);
		b1 = fits[type][counter_a-1][3] + (fits[type][counter_a][3] - fits[type][counter_a-1][3])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]);
		c1 = fits[type][counter_a-1][4] + (fits[type][counter_a][4] - fits[type][counter_a-1][4])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]);
		d1 = fits[type][counter_a-1][5] + (fits[type][counter_a][5] - fits[type][counter_a-1][5])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]);
		e1 = fits[type][counter_a-1][6] + (fits[type][counter_a][6] - fits[type][counter_a-1][6])/(gamma_a[counter_a] - gamma_a[counter_a-1])*(a - gamma_a[counter_a-1]);
		return dof[particle_index]*pow(10.,a1*(2.*M*E*G) + b1 + c1*cos(e1*2.*E*M*G) + d1*sin(e1*2.*E*M*G))/(2.*pi);
	}
}

void instantaneous_primary_spectrum(double **instantaneous_primary_spectra,double **BH_masses,double **BH_spins,double **spec_table,double *energies,
	double ***gammas,double *gamma_a,double *gamma_x,double ***fits,double *dof,double *spins,double *masses_primary,struct param *parameters){
	// This function computes the instantaneous primary spectra of a distribution of BHs
	// and stores is it in the array instantaneous_primary_spectra[][].
	
	int **counters_a = (int **)malloc(parameters->BHnumber*sizeof(int *)); // contains the running position in the tabulated spins
	int ***counters_x = (int ***)malloc(parameters->BHnumber*sizeof(int **)); // contains the running position in the tabulated energies
	for(int i = 0;i<parameters->BHnumber;i++){
		counters_a[i] = (int *)malloc(parameters->anumber*sizeof(int));
		for(int j = 0;j<parameters->anumber;j++){
			counters_a[i][j] = 0;
			while(counters_a[i][j] < parameters->nb_gamma_a - 1 && BH_spins[i][j] >= gamma_a[counters_a[i][j]]){ // finding the nearest spin parameter
				counters_a[i][j]++;
			}
		}
		counters_x[i] = (int **)malloc(parameters->anumber*sizeof(int *));
		for(int j = 0;j<parameters->anumber;j++){
			counters_x[i][j] = (int *)malloc(parameters->Enumber*sizeof(int));
			counters_x[i][j][0] = 0;
			for(int k = 0;k<parameters->Enumber;k++){
				if(k>0){
					counters_x[i][j][k] = counters_x[i][j][k-1];
				}
				while(counters_x[i][j][k] < parameters->nb_gamma_x - 1 && energies[k]*2.*BH_masses[i][j]*G >= gamma_x[counters_x[i][j][k]]){ // finding the nearest omega*r_H
					counters_x[i][j][k]++;
				}
			}
		}
	}
	
	int *compute = (int *)malloc((parameters->particle_number+parameters->grav)*sizeof(int)); // this table decides whether the corresponding spectra are computed or not
	compute[0] = 1; // photons
	compute[1] = 1; // gluons
	compute[2] = 1; // higgs boson
	compute[3] = 1; // W+- boson
	compute[4] = 1; // Z0 boson
	compute[5] = 1; // neutrinos
	compute[6] = 1; // electron
	compute[7] = 1; // muon
	compute[8] = 1; // tau
	compute[9] = 1; // up quark
	compute[10] = 1; // down quark
	compute[11] = 1; // charm quark
	compute[12] = 1; // strange quark
	compute[13] = 1; // top quark
	compute[14] = 1; // bottom quark
	if(parameters->grav){
		compute[15] = 1; // graviton
	}
	
	if(parameters->full_output){
		printf("\n\n\t compute:\t");
		fflush(stdout);
	}
	for(int i = 0;i<parameters->particle_number+parameters->grav;i++){
/*#if defined(_OPENMP)
#pragma omp parallel for
#endif*/
		for(int j = 0;j<parameters->Enumber;j++){
			instantaneous_primary_spectra[i][j] = 0.;
			if(compute[i]){
				for(int k = 0;k<parameters->BHnumber;k++){
					for(int l = 0;l<parameters->anumber;l++){
						instantaneous_primary_spectra[i][j] = instantaneous_primary_spectra[i][j] + spec_table[k][l]*dNdtdE(energies[j],BH_masses[k][l],BH_spins[k][l],i,gammas,gamma_a,gamma_x,fits,dof,spins,masses_primary,counters_a[k][l],counters_x[k][l][j],parameters);
					}
				}
			}
		}
		if(parameters->full_output){ // prints the current state of computation
			printf("%i/%i\t",i+1,parameters->particle_number+parameters->grav);
			fflush(stdout);
		}
	}
	
	if(parameters->full_output){
		printf("\n\n");
		fflush(stdout);
	}
	free2D_int(counters_a,parameters->BHnumber);
	free3D_int(counters_x,parameters->BHnumber,parameters->anumber);
	return;
}

void write_instantaneous_primary_spectra(double **instantaneous_primary_spectra,double *energies,struct param *parameters){
	// This function writes the instantaneous Hawking primary spectrum of a distribution of BH,
	// contained in the array instantaneous_primary_spectra[][], into the file 'instantaneous_primary_spectra.txt'.
	
	char file_name[500];
	sprintf(file_name,"./results/%s/instantaneous_primary_spectra.txt",parameters->destination_folder);
	FILE *file = fopen(file_name,"w");
	if(!file){
		printf("\n\t [write_instantaneous_primary_spectra] : ERROR COULD NOT OPEN FILE '%s' !\n",file_name);
		fflush(stdout);
		exit(0);
	}
	rewind(file);
	fprintf(file,"Hawking primary spectra for each particle types.\n");
	switch(parameters->grav){
		case 0:{
			fprintf(file,"%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s\n","energy/particle","photon","gluons","higgs","W+-","Z0","neutrinos","electron","muon","tau","up","down","charm","strange","top","bottom");
			break;
		}
		case 1:{
			fprintf(file,"%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s%15s\n","energy/particle","photon","gluons","higgs","W+-","Z0","neutrinos","electron","muon","tau","up","down","charm","strange","top","bottom","graviton");
			break;
		}
		default:{
			printf("\n\t [write_instantaneous_primary_spectra] : ERROR WRONG GRAVITON CHOICE !\n");
			fflush(stdout);
			exit(0);
			break;
		}
	}
	for(int j = 0;j<parameters->Enumber;j++){
		fprintf(file,"%15.5e",energies[j]);
		for(int i = 0;i<parameters->particle_number+parameters->grav;i++){
			fprintf(file,"%15.5e",instantaneous_primary_spectra[i][j]*rate_conversion); // conversion from GeV to CGS units
		}
		fprintf(file,"\n");
	}
	fclose(file);
	return;
}
